/* This file is part of the iox-ili project.
 * For more information, please see <http://www.eisenhutinformatik.ch/iox-ili/>.
 *
 * Copyright (c) 2006 Eisenhut Informatik AG
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included 
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */
package ch.interlis.ilirepository.impl;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigInteger;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

import ch.ehi.basics.logging.EhiLogger;
import ch.ehi.iox.ilisite.IliRepository09.RepositoryIndex.ModelMetadata;
import ch.ehi.iox.ilisite.IliRepository09.RepositoryIndex.ModelMetadata_SchemaLanguage;
import ch.interlis.ili2c.modelscan.IliFile;
import ch.interlis.ili2c.modelscan.IliModel;
import ch.interlis.ilirepository.IliFiles;
import ch.interlis.ilirepository.IliSite;
import ch.interlis.ilirepository.IliManager;
import ch.interlis.iox.IoxException;

/** Handles access to a file in a repository.
 * Hides the caching, downloading of files.
 * @author ceis
 */
public class RepositoryAccess {
	private HashMap reposIliFiles=new HashMap();
	private HashMap reposIliSite=new HashMap();
	private long metaMaxTTL=86400000L; // max time to live in ms for a file in the cache
	private File localCache=new File(System.getProperty("user.home"),".ilicache");

	/** Sets the local folder to be used for caching remote files.
	 */
	public void setCache(File localdir){
		localCache=localdir;
	}
	/** Gets the list of ili-files for a given repository.
	 * @param uri local or remote repository
	 * @return list of ili-files or null if no such repository or offline.
	 */
	public IliFiles getIliFiles(String uri)
	{
		if(!reposIliFiles.containsKey(uri)){
			if(isLegacyDir(uri)){
				// scan directory
				try {
					EhiLogger.traceState("scan ili-files in folder <"+uri+">...");
					HashSet files=ch.interlis.ili2c.ModelScan.scanIliFileDir(new File(uri), null);
					Iterator filei=files.iterator();
					IliFiles iliFiles=new IliFiles();
					while(filei.hasNext()){
						iliFiles.addFile((IliFile)filei.next());
					}
					reposIliFiles.put(uri, iliFiles);
				} catch (IOException e) {
					throw new IllegalArgumentException(e);
				}
			}else{
				IliFiles iliFiles;
				try {
					EhiLogger.traceState("read "+IliManager.ILIMODELS_XML+" from <"+uri+">...");
					iliFiles = readIliModelsXml(uri);
				} catch (RepositoryAccessException e) {
					handleRepositoryAccessException(e,uri);
					iliFiles=null;
				}
				reposIliFiles.put(uri, iliFiles);
			}
		}
		IliFiles iliFiles=(IliFiles)reposIliFiles.get(uri);
		return iliFiles;
	}
	/** Gets the site metadata for a given repository.
	 * @param uri local or remote repository
	 * @return site metadata or null if no such repository or offline.
	 */
	public IliSite getIliSite(String uri) 
	{
		if(!reposIliSite.containsKey(uri)){
			IliSite iliSite;
			try {
				EhiLogger.traceState("read "+IliManager.ILISITE_XML+" from <"+uri+">...");
				iliSite = readIliSiteXml(uri);
			} catch (RepositoryAccessException e) {
				handleRepositoryAccessException(e,uri);
				iliSite=null;
			}
			reposIliSite.put(uri, iliSite);
		}
		IliSite iliSite=(IliSite)reposIliSite.get(uri);
		return iliSite;
	}
	
	private void handleRepositoryAccessException(RepositoryAccessException e,String uri)
	{
		if(reposIliSite.containsKey(uri) && reposIliFiles.containsKey(uri)){
			return;
		}
		EhiLogger.logAdaption("repository <"+uri+"> ignored; "+e.getCause());
		if(!reposIliSite.containsKey(uri)){
			reposIliSite.put(uri, null);
		}
		if(reposIliFiles.containsKey(uri)){
			reposIliFiles.put(uri, null);
		}
	}
	private boolean isLegacyDir(String uri)
	{
		String urilc=uri.toLowerCase();
		if(urilc.startsWith("http:")){
			return false;
		}
		File file=new File(uri,IliManager.ILIMODELS_XML);
		if(file.exists()){
			return false;
		}
		return true;
	}
	/** read an ilimodels.xml file
	 * 
	 * @param uri uri of the repository without ilimodels.xml
	 * @return null if the repository doesn't exist
	 */
	private IliFiles readIliModelsXml(String uri)
	throws RepositoryAccessException
	{
		File file=getLocalFileLocation(uri,IliManager.ILIMODELS_XML,metaMaxTTL,null);
		if(file==null){
			return null;
		}
		// read file
		HashMap models=new HashMap(); // map<String modelName,array<ModelMetadata>>
		ch.interlis.iom_j.xtf.XtfReader reader=null;
		try {
			reader=new ch.interlis.iom_j.xtf.XtfReader(file);
			reader.getFactory().registerFactory(ch.ehi.iox.ilisite.ILIREPOSITORY09.getIoxFactory());
			ch.interlis.iox.IoxEvent event=null;
			do{
				 event=reader.read();
				 if(event instanceof ch.interlis.iox.ObjectEvent){
					 ch.interlis.iom.IomObject iomObj=((ch.interlis.iox.ObjectEvent)event).getIomObject();
					 if(iomObj instanceof ch.ehi.iox.ilisite.IliRepository09.RepositoryIndex.ModelMetadata){
						 ch.ehi.iox.ilisite.IliRepository09.RepositoryIndex.ModelMetadata model=(ch.ehi.iox.ilisite.IliRepository09.RepositoryIndex.ModelMetadata)iomObj;
						 String name=model.getName();
						 if(!models.containsKey(name)){
							 models.put(name,new ArrayList());
						 }
						 ArrayList versions=(ArrayList)models.get(name);
						 versions.add(model);
					 }
				 }
			}while(!(event instanceof ch.interlis.iox.EndTransferEvent));
		} catch (IoxException e) {
			throw new IllegalStateException(e);
		}finally{
			if(reader!=null){
				try {
					reader.close();
				} catch (IoxException e) {
					throw new IllegalStateException(e);
				}
				reader=null;
			}
		}
		IliFiles ret=new IliFiles();
		HashMap files=new HashMap();
		Iterator modeli=models.keySet().iterator();
		while(modeli.hasNext()){
			String name=(String)modeli.next();
			ArrayList modelVersions=(ArrayList)models.get(name);
			ModelMetadata_SchemaLanguage cslv[]=new ModelMetadata_SchemaLanguage[]{ModelMetadata_SchemaLanguage.ili1,ModelMetadata_SchemaLanguage.ili2_2,ModelMetadata_SchemaLanguage.ili2_3};  
			for(int csli=0;csli<cslv.length;csli++){
				ModelMetadata model=getLatestVersion(modelVersions,cslv[csli]);
				if(model==null){
					continue;
				}
				String filename=model.getFile();
				IliFile ilifile=null;
				if(!files.containsKey(filename)){
					ilifile=new IliFile();
					ilifile.setPath(filename);
					ilifile.setRepositoryUri(uri);
					files.put(filename, ilifile);
					ret.addFile(ilifile);
				}else{
					ilifile=(IliFile)files.get(filename);
				}
				IliModel iliModel=new IliModel();
				ModelMetadata_SchemaLanguage csl=model.getSchemaLanguage();
				if(csl==ModelMetadata_SchemaLanguage.ili1){
					iliModel.setIliVersion(1.0);
				}else if(csl==ModelMetadata_SchemaLanguage.ili2_2){
					iliModel.setIliVersion(2.2);
				}else if(csl==ModelMetadata_SchemaLanguage.ili2_3){
					iliModel.setIliVersion(2.3);
				}
				iliModel.setName(model.getName());
				ch.ehi.iox.ilisite.IliRepository09.ModelName_ deps[]=model.getdependsOnModel();
				for(int depi=0;depi<deps.length;depi++){
					iliModel.addDepenedency(deps[depi].getvalue());
				}
				ilifile.addModel(iliModel);
			}
		}
		return ret;
	}
	/** read an ilisite.xml file
	 * 
	 * @param uri uri of the repository (without ilisite.xml)
	 * @return null if the repository doesn't exist
	 */
	private IliSite readIliSiteXml(String uri)
	throws RepositoryAccessException
	{
		// read it (local location or from cache)
		File file=getLocalFileLocation(uri,IliManager.ILISITE_XML,metaMaxTTL,null);
		if(file==null){
			return null;
		}
		// read file
		ch.interlis.iom_j.xtf.XtfReader reader=null;
		try {
			reader=new ch.interlis.iom_j.xtf.XtfReader(file);
			reader.getFactory().registerFactory(ch.ehi.iox.ilisite.ILISITE09.getIoxFactory());
			ch.interlis.iox.IoxEvent event=null;
			do{
				 event=reader.read();
				 if(event instanceof ch.interlis.iox.ObjectEvent){
					 ch.interlis.iom.IomObject iomObj=((ch.interlis.iox.ObjectEvent)event).getIomObject();
					 if(iomObj instanceof ch.ehi.iox.ilisite.IliSite09.SiteMetadata.Site){
						 ch.ehi.iox.ilisite.IliSite09.SiteMetadata.Site site=(ch.ehi.iox.ilisite.IliSite09.SiteMetadata.Site)iomObj;
						 IliSite ret=new IliSite();
						 ch.ehi.iox.ilisite.IliSite09.RepositoryLocation_ parents[]=site.getparentSite();
						 for(int parenti=0;parenti<parents.length;parenti++){
							 ret.addParentSite(parents[parenti].getvalue());
						 }
						 ch.ehi.iox.ilisite.IliSite09.RepositoryLocation_ subsidiaries[]=site.getsubsidiarySite();
						 for(int subsidiaryi=0;subsidiaryi<subsidiaries.length;subsidiaryi++){
							 ret.addSubsidiarySite(subsidiaries[subsidiaryi].getvalue());
						 }
						 return ret;
					 }
				 }
			}while(!(event instanceof ch.interlis.iox.EndTransferEvent));
		} catch (IoxException e) {
			throw new IllegalStateException(e);
		}finally{
			if(reader!=null){
				try {
					reader.close();
				} catch (IoxException e) {
					throw new IllegalStateException(e);
				}
				reader=null;
			}
		}
		return null;
	}
	private ModelMetadata getLatestVersion(ArrayList modelVersions1,ModelMetadata_SchemaLanguage csl)
	{
		ArrayList modelVersions=new ArrayList();
		ch.ehi.iox.ilisite.IliRepository09.RepositoryIndex.ModelMetadata model=null;

		// remove different csl
		for(Iterator i=modelVersions1.iterator();i.hasNext();){
			model=(ch.ehi.iox.ilisite.IliRepository09.RepositoryIndex.ModelMetadata)i.next();
			if(model.getSchemaLanguage().equals(csl)){
				modelVersions.add(model);
			}
		}
		if(modelVersions.isEmpty()){
			// not a model with given csl
			return null;
		}
		
		// find first version and eliminate other versions without duplicate
		ch.ehi.iox.ilisite.IliRepository09.RepositoryIndex.ModelMetadata first=null;
		for(Iterator i=modelVersions.iterator();i.hasNext();){
			model=(ch.ehi.iox.ilisite.IliRepository09.RepositoryIndex.ModelMetadata)i.next();
			if(model.getprecursorVersion()==null){
				if(first==null){
					if(!model.getbrowseOnly()){
						first=model;
					}
					i.remove();
				}else{
					if(!model.getbrowseOnly()){
						EhiLogger.logAdaption("model "+model.getName()+": duplicate version without precursor; version "+model.getVersion()+" ignored");
					}
					i.remove();
				}
			}
		}
		if(first==null){
			EhiLogger.logAdaption("model "+model.getName()+": no version without precursor; model ignored");
			return null;
		}
		ch.ehi.iox.ilisite.IliRepository09.RepositoryIndex.ModelMetadata lastVersion=first;
		ch.ehi.iox.ilisite.IliRepository09.RepositoryIndex.ModelMetadata current=first;
		while(!modelVersions.isEmpty()){
			// find next version
			ch.ehi.iox.ilisite.IliRepository09.RepositoryIndex.ModelMetadata next=null;
			for(Iterator i=modelVersions.iterator();i.hasNext();){
				model=(ch.ehi.iox.ilisite.IliRepository09.RepositoryIndex.ModelMetadata)i.next();
				if(current.getVersion().equals(model.getprecursorVersion())){
					if(next==null){
						// next version found
						next=model;
						i.remove();
					}else{
						// another next version found; ignore it
						EhiLogger.logAdaption("model "+model.getName()+": duplicate version with same precursor; version "+model.getVersion()+" ignored");
						i.remove();
					}
				}
			}
			if(next==null){
				// no next version found
				break;
			}
			current=next;
			if(!current.getbrowseOnly()){
				lastVersion=current;
			}
		}
		if(!modelVersions.isEmpty()){
			EhiLogger.logAdaption("model "+lastVersion.getName()+": broken version chain; use version "+lastVersion.getVersion()+"; others ignored");
		}
		return lastVersion;
	}
	/** Gets access to a file.
	 * If the file is in a remote repository, it will be downloaded to the cache and path to the version in the cache is returned.
	 * If the file is already in the cache, the repository will be checked for never version depending on age of cached file and md5 digest.
	 * @param uri remote or local repository
	 * @param filename file/path to download from repository. If null, uri includes the file name.
	 * @param maxTTL max age of file in cache before downloading it again.
	 * @param md5 digest of remote file or null. If cached version has a different digest, download remote file again.
	 * @return local file or null if it doesn't exist in remote repository
	 * @throws RepositoryAccessException
	 */
	public File getLocalFileLocation(String uri,String filename,long maxTTL,String md5)
	throws RepositoryAccessException
{
	if(uri==null){
		if(filename!=null){
			return new File(filename);
		}
		return null;
	}
	// if http
	String urilc=uri.toLowerCase();
	if(urilc.startsWith("http:")){
		// translate uri to lcation in cache
		StringBuffer localFileName=new StringBuffer();
		{
			String urib=uri.substring("http:".length()); 
			// escape characters
			// win < > : " / \ | ? * %
			for(int i=0;i<urib.length();i++){
				char c=urib.charAt(i);
				if(c=='<'
						|| c=='>'
						|| c==':'
						|| c=='"'
						|| c=='\\'
						|| c=='|'
						|| c=='?'
						|| c=='*'
						|| c=='%'
						|| c=='&'){
					localFileName.append('&');
					String str=Integer.toHexString(c);
					localFileName.append("0000".substring(str.length())+str);
				}else{
					localFileName.append(c);
				}
			}
		}
		File ret=new File(localCache,localFileName.toString());
		if(filename!=null){
			ret=new File(ret,filename);
		}
		boolean fetchFromServer=true;
		if(!ret.exists()){
			fetchFromServer=true;
		}else{
			fetchFromServer=false;
			if(!fetchFromServer && (maxTTL==0 || ret.lastModified()+maxTTL<System.currentTimeMillis())){
				fetchFromServer=true;
			}
			if(!fetchFromServer && md5!=null && !calcMD5(ret).equals(md5)){
				fetchFromServer=true;
			}
		}
		// if not in cache
		if(fetchFromServer){
			// fetch from http server (handle redirects)
			java.net.URL url=null;
			try {
				url=new java.net.URI(uri).toURL();
				if(filename!=null){
					// incomplete url of directory?
					if(!uri.endsWith("/")){
						// fix it
						url=new java.net.URI(uri+"/").toURL();
					}
					url=new java.net.URL(url,filename);
				}
				EhiLogger.traceState("fetching <"+url+"> ...");
			} catch (MalformedURLException e) {
				throw new RepositoryAccessException(e);
			} catch (URISyntaxException e) {
				throw new RepositoryAccessException(e);
			}
			java.net.URLConnection conn=null;
			try {
				conn = url.openConnection();
			} catch (IOException e) {
				throw new RepositoryAccessException(e);
			}
			java.io.BufferedInputStream in=null;
			java.io.OutputStream fos=null;
			try{
				try {
					in=new java.io.BufferedInputStream(conn.getInputStream());
				} catch (FileNotFoundException e) {
					return null;
				//} catch (java.net.UnknownHostException e) {
				//	return null;
				} catch (IOException e) {
					throw new RepositoryAccessException(e);
				}
				// create directory
				java.io.File dir=ret.getParentFile();
				if(!dir.exists()){
					boolean created=dir.mkdirs();
					if(!created){
						throw new IllegalArgumentException("failed to create folder "+uri);
					}
				}
				// save to cache
				try {
					fos = new java.io.BufferedOutputStream(new java.io.FileOutputStream(ret));
				} catch (FileNotFoundException e) {
					throw new IllegalArgumentException("failed to create file in cache",e);
				}
			    try {
					byte[] buf = new byte[1024];
					int i = 0;
					while ((i = in.read(buf)) != -1) {
					    fos.write(buf, 0, i);
					}
				} catch (IOException e) {
					throw new IllegalArgumentException("failed to save downloaded file to cache",e);
				}
			}finally{
				if(in!=null){
					try {
						in.close();
					} catch (IOException e) {
						EhiLogger.logError(e);
					}
					in=null;
				}
				if(fos!=null){
					try {
						fos.close();
					} catch (IOException e) {
						EhiLogger.logError(e);
					}
					fos=null;
				}
			}
		}
		return ret;
	}
	// uri is a local file 
	File ret=new File(uri);
	if(filename!=null){
		ret=new File(ret,filename);
	}
	if(!ret.exists()){
		return null;
	}
	return ret;
}
	/** Calculates the digest of a local file.
	 * @param file local file
	 * @return md5 digest of local file content.
	 */
	public static String calcMD5(File file)
	{
		java.security.MessageDigest md;
		try {
			md = java.security.MessageDigest.getInstance("MD5");
		} catch (NoSuchAlgorithmException e) {
			throw new IllegalArgumentException(e);
		}
		java.io.InputStream is;
		try {
			is = new java.io.FileInputStream(file);
		} catch (FileNotFoundException e) {
			throw new IllegalArgumentException(e);
		}
		String signature = null;
		try {
			byte[] buffer = new byte[1024];
		     int numRead;
		     do {
		      numRead = is.read(buffer);
		      if (numRead > 0) {
		        md.update(buffer, 0, numRead);
		        }
		      } while (numRead != -1);
			signature = new java.math.BigInteger(1,md.digest()).toString(16);
		}catch(IOException e){
			EhiLogger.logError(e);
		}finally {
			try {
				is.close();
			} catch (IOException e) {
				EhiLogger.logError(e);
			}
		}
		return signature;
	}
}
